import logging

from bottle import request, app


class DuplicateCommandError(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "Command already defined: %s" % self.name


class DuplicatePluginError(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "Plugin already defined: %s" % self.name


class BaseCommand(object):
    def __init__(self, instance, method):
        self.instance = instance
        self.method = method
        self.help = method.__doc__

    def execute(self, *args, **kwargs):
        return self.method(self.instance, *args, **kwargs)


class PluginCommand(BaseCommand):
    def __init__(self, instance, method):
        super(PluginCommand, self).__init__(instance, method)
        self.acl = method.acl
        self.admin_only = method.admin_only


class WebhookCommand(BaseCommand):
    def __init__(self, instance, method, form_params):
        super(WebhookCommand, self).__init__(instance, method)
        self.form_params = form_params

    def execute(self):
        args = {}
        form_params = self.form_params
        if isinstance(self.form_params, basestring):
            form_params = [self.form_params]
        for p in form_params:
            args[p] = request.forms[p]
        return self.method(self.instance, **args)


class MessageDispatcher(object):
    def __init__(self):
        self.log = logging.getLogger(__name__)
        self.commands = {}

    def push(self, message):
        """
        Takes a SlackEvent, parses it for a command, and runs against registered plugin
        """
        args = self._parse_message(message)
        cmd = args[0]
        msg_args = args[1:]
        self.log.debug("Received command %s with args %s", cmd, msg_args)
        if cmd in self.commands:
            if message.user is None:
                self.log.debug("Discarded message with no originating user: %s", message)
                return None, None
            sender = message.user.username
            if message.channel is not None:
                sender = "#%s/%s" % (message.channel.channel, sender)
            self.log.info("Received from %s: %s, args %s", sender, cmd, msg_args)
            f = self._get_command(cmd, message.user)
            if f:
                return cmd, f.execute(message, msg_args)
            return '_unauthorized_', "Sorry, you are not authorized to run %s" % cmd
        return None, None

    def _parse_message(self, message):
        args = message.text.split(' ')
        return args

    def register_plugin(self, plugin):
        """Registers a plugin and commands with the dispatcher for push()"""
        self.log.info("Registering plugin %s", plugin.__class__.__name__)
        self._register_commands(plugin)
        plugin.on_load()

    def _register_commands(self, plugin):
        for name, method in plugin.__class__.__dict__.iteritems():
            if hasattr(method, 'is_cmd'):
                commands = [name]
                if method.aliases is not None:
                    aliases = method.aliases
                    if isinstance(method.aliases, basestring):
                        aliases = [method.aliases]
                    for alias in aliases:
                        commands.append(alias)

                for cmd_name in commands:
                    cmd = '!' + cmd_name
                    if cmd in self.commands:
                        raise DuplicateCommandError(cmd_name)
                    self.log.info("Registered command %s", plugin.__class__.__name__ + '.' + cmd_name)
                    self.commands[cmd] = PluginCommand(plugin, method)
            elif hasattr(method, 'is_webhook'):
                self.log.info("Registered webhook %s", plugin.__class__.__name__ + '.' + name)
                webhook = WebhookCommand(plugin, method, method.form_params)
                app().route(method.route, 'POST', webhook.execute)

    def _get_command(self, cmd, user):
        can_run_cmd = True
        if hasattr(self, 'auth_manager'):
            can_run_cmd = self.auth_manager.admin_check(self.commands[cmd], user)
            if can_run_cmd:
                can_run_cmd = self.auth_manager.acl_check(self.commands[cmd], user)
        if not can_run_cmd:
            self.log.info("User %s is not authorized to run %s", user.username, cmd)
            return None
        return self.commands[cmd]



import logging

from datetime import datetime
from slackclient import SlackClient
from time import sleep

from dispatcher import MessageDispatcher
from slackminion.plugin.manager import PluginManager
from slack import SlackEvent, SlackChannel, SlackUser
from webserver import Webserver


class NotSetupError(Exception):
    def __str__(self):
        return "Bot not setup.  Please run start() before run()."


def eventhandler(*args, **kwargs):
    """
    Decorator.  Marks a function as a receiver for the specified slack event(s).

    * events - String or list of events to handle
    """
    def wrapper(func):
        if isinstance(kwargs['events'], basestring):
            kwargs['events'] = [kwargs['events']]
        func.is_eventhandler = True
        func.events = kwargs['events']
        return func
    return wrapper


class Bot(object):
    def __init__(self, config, test_mode=False):
        self.always_send_dm = []
        self.config = config
        self.dispatcher = MessageDispatcher()
        self.event_handlers = {}
        self.is_setup = False
        self.log = logging.getLogger(__name__)
        self.plugins = PluginManager(self, test_mode)
        self.runnable = True
        self.sc = None
        self.webserver = None
        self.test_mode = test_mode

        if self.test_mode:
            self.metrics = {
                'startup_time': 0
            }

        from slackminion import version
        self.version = version
        try:
            from slackminion import commit
            self.commit = commit
        except ImportError:
            self.commit = "HEAD"

    def start(self):
        """Initializes the bot, plugins, and everything."""
        if self.test_mode:
            bot_start_time = datetime.now()
        self.plugins.load()
        self.plugins.load_state()
        self._find_event_handlers()
        self.sc = SlackClient(self.config['slack_token'])
        self.webserver = Webserver(self.config['webserver']['host'], self.config['webserver']['port'])

        self.always_send_dm = ['_unauthorized_']
        if 'always_send_dm' in self.config:
            self.always_send_dm.extend(map(lambda x: '!' + x, self.config['always_send_dm']))

        # Rocket is very noisy at debug
        logging.getLogger('Rocket.Errors.ThreadPool').setLevel(logging.INFO)

        self.is_setup = True
        if self.test_mode:
            self.metrics['startup_time'] = (datetime.now() - bot_start_time).total_seconds() * 1000.0

    def _find_event_handlers(self):
        for name, method in self.__class__.__dict__.iteritems():
            if hasattr(method, 'is_eventhandler'):
                for event in method.events:
                    self.event_handlers[event] = method

    def run(self):
        """Connects to slack and enters the main loop."""
        # Fail out if setup wasn't run
        if not self.is_setup:
            raise NotSetupError

        if not self.sc.rtm_connect():
            return False

        # Start the web server
        self.webserver.start()
        try:
            while self.runnable:
                # Get all waiting events - this always returns a list
                events = self.sc.rtm_read()
                for e in events:
                    self._handle_event(e)
                sleep(0.1)
        except KeyboardInterrupt:
            # On ctrl-c, just exit
            pass
        except:
            self.log.exception('Unhandled exception')

    def stop(self):
        """Does cleanup of bot and plugins."""
        if self.webserver is not None:
            self.webserver.stop()
        if not self.test_mode:
            self.plugins.save_state()

    def send_message(self, channel, text):
        """
        Sends a message to the specified channel

        * channel - The channel to send to.  This can be a SlackChannel object, a channel id, or a channel name (without the #)
        * text - String to send
        """
        # This doesn't want the # in the channel name
        if isinstance(channel, SlackChannel):
            channel = channel.channelid
        self.log.debug("Trying to send to %s: %s", channel, text)
        self.sc.rtm_send_message(channel, text)

    def send_im(self, user, text):
        """
        Sends a message to a user as an IM

        * user - The user to send to.  This can be a SlackUser object, a user id, or the username (without the @)
        * text - String to send
        """
        if isinstance(user, SlackUser):
            user = user.userid
        channelid = self._find_im_channel(user)
        self.send_message(channelid, text)

    def _find_im_channel(self, user):
        resp = self.sc.api_call('im.list')
        channels = filter(lambda x: x['user'] == user, resp['ims'])
        if len(channels) > 0:
            return channels[0]['id']
        resp = self.sc.api_call('im.open', user=user)
        return resp['channel']['id']

    def _load_user_rights(self, user):
        if user is not None:
            if 'bot_admins' in self.config:
                if user.username in self.config['bot_admins']:
                    user.is_admin = True

    def _handle_event(self, event):
        if 'type' not in event:
            # This is likely a notification that the bot was mentioned
            self.log.debug("Received odd event: %s", event)
            return
        e = SlackEvent(sc=self.sc, **event)
        self.log.debug("Received event type: %s", e.type)

        if e.user is not None:
            if hasattr(self, 'user_manager'):
                user = self.user_manager.get(e.user.userid)
                if user is None:
                    user = self.user_manager.set(e.user)
                e.user = user

        if e.type in self.event_handlers:
            self.event_handlers[e.type](self, e)

    @eventhandler(events='message')
    def _event_message(self, msg):
        self.log.debug("Message.message: %s: %s: %s", msg.channel, msg.user, msg.__dict__)

        # The user manager should load rights when a user is added
        if not hasattr(self, 'user_manager'):
            self._load_user_rights(msg.user)
        try:
            cmd, output = self.dispatcher.push(msg)
        except:
            self.log.exception('Unhandled exception')
            return
        self.log.debug("Output from dispatcher: %s", output)
        if output:
            if cmd in self.always_send_dm:
                self.send_im(msg.user, output)
            else:
                self.send_message(msg.channel, output)

    @eventhandler(events='error')
    def _event_error(self, msg):
        self.log.error("Received an error response from Slack: %s", msg.__dict__)



import argparse
import logging
import signal
import sys
import yaml

from slackminion.bot import Bot


def main():
    def sigterm_handler(signum, frame):
        bot.runnable = False

    p = argparse.ArgumentParser()
    p.add_argument('--config', action='store', default='config.yaml', help='Specify a config file (default: config.yaml)')
    p.add_argument('--test', action='store_true', help='Load plugins and exit')
    args = p.parse_args()

    with open(args.config, 'rb') as f:
        config = yaml.load(f)

    level = logging.DEBUG if config['debug'] else logging.INFO
    logging.basicConfig(level=level, format='%(asctime)s %(name)s %(levelname)s: %(message)s')

    bot = Bot(config, args.test)
    bot.start()
    if not args.test:
        signal.signal(signal.SIGTERM, sigterm_handler)
        bot.run()
    bot.stop()

    if args.test:
        test_passed = True
        output = ["Bot Test Results"]
        metrics = bot.plugins.metrics
        output.append("Plugins Loaded")
        for p in bot.plugins.plugins:
            output.append("{:<40} {:>7.03f} ms".format(p.__class__.__name__, metrics['load_times'][p.__class__.__name__]))
        if len(metrics['plugins_failed']) > 0:
            output.append("")
            output.append("Plugins failed to load")
            for p in metrics['plugins_failed']:
                output.append(p)
        output.append("")
        output.append("Bot startup time: %.03f ms" % bot.metrics['startup_time'])
        output.append("Plugins: %d total, %d loaded, %d failed" % (metrics['plugins_total'], metrics['plugins_loaded'], len(metrics['plugins_failed'])))
        if metrics['plugins_total'] != metrics['plugins_loaded']:
            output.append("")
            output.append("=== Bot Failed Startup Tests ===")
            test_passed = False
        logging.getLogger().info('\n'.join(output))

        if not test_passed:
            sys.exit(1)

if __name__ == "__main__":
    main()



from operator import itemgetter

from slackminion.plugin import cmd
from slackminion.plugin.base import BasePlugin


class Core(BasePlugin):

    @cmd()
    def help(self, msg, args):
        """Displays help for each command"""
        output = []
        if len(args) == 0:
            commands = sorted(self._bot.dispatcher.commands.items(), key=itemgetter(0))
            # Filter commands if auth is enabled, hide_admin_commands is enabled, and user is not admin
            if hasattr(self._bot.dispatcher, 'auth_manager') and \
                    'hide_admin_commands' in self._bot.config and \
                    self._bot.config['hide_admin_commands'] is True and \
                    not getattr(msg.user, 'is_admin', False):
                commands = filter(lambda x: x[1].admin_only is False, commands)
            for name, cmd in commands:
                helpstr = cmd.help
                if '.' in helpstr:
                    helpstr = helpstr[0:helpstr.find('.')+1]
                if helpstr is None:
                    helpstr = "No description provided."
                output.append("*%s*: %s" % (name, helpstr))
        else:
            name = '!' + args[0]
            if name not in self._bot.dispatcher.commands:
                output = ['No such command: %s' % name]
            else:
                helpstr = self._bot.dispatcher.commands[name].help
                if helpstr is None:
                    helpstr = "No description provided."
                output = [helpstr]
        return '\n'.join(output)

    @cmd(admin_only=True)
    def save(self, msg, args):
        """Causes the bot to write its current state to backend."""
        self.send_message(msg.channel, "Saving current state...")
        self._bot.plugins.save_state()
        self.send_message(msg.channel, "Done.")

    @cmd(admin_only=True)
    def shutdown(self, msg, args):
        """Causes the bot to gracefully shutdown."""
        self.log.info("Received shutdown from %s", msg.user.username)
        self._bot.runnable = False
        return "Shutting down..."

    @cmd()
    def whoami(self, msg, args):
        """Prints information about the user and bot version."""
        output = ["Hello %s" % msg.user]
        if hasattr(self._bot.dispatcher, 'auth_manager') and msg.user.is_admin is True:
            output.append("You are a *bot admin*.")
        output.append("Bot version: %s-%s" % (self._bot.version, self._bot.commit))
        return '\n'.join(output)



import redis

from . import BaseStateHandler


class RedisStateHandler(BaseStateHandler):
    def load_state(self):
        r, prefix = self._get_redis_info()
        state = r.get(prefix + ':state')
        return state

    def save_state(self, state):
        r, prefix = self._get_redis_info()
        r.set(prefix + ':state', state)

    def _get_redis_info(self):
        r = redis.StrictRedis(host=self.config['host'], port=self.config['port'])
        prefix = 'slack'
        if 'prefix' in self.config:
            prefix = self.config['prefix']
        return r, prefix



from ..plugin import cmd, webhook
from slackminion.plugin.base import BasePlugin


class TestPlugin(BasePlugin):

    @cmd()
    def echo(self, msg, args):
        """Simply repeats whatever is said."""
        self.log.debug("Received args: %s", args)
        return ' '.join(args)

    @cmd()
    def xyzzy(self, msg, args):
        """Nothing happens."""
        return "Nothing happens for %s" % msg.user

    @cmd()
    def alert(self, msg, args):
        """Alert everyone."""
        self.send_message(self.config['channel'], '<!here>: something important is going to happen!')
        return None

    @webhook('/echo', form_params='foo')
    def web_echo(self, foo):
        self.send_message(self.config['channel'], foo)

    @cmd()
    def sleep(self, msg, args):
        """Sleep for a bit, then print a message."""
        self.start_timer(5, self._sleep_func)

    @cmd()
    def sleep2(self, msg, args):
        """Sleep for a bit, then echo the message back"""
        self.start_timer(5, self._sleep_func2, msg.channel, ' '.join(args))

    def _sleep_func(self):
        self.send_message(self.config['channel'], 'Slept for a bit')

    def _sleep_func2(self, channel, text):
        self.send_message(channel, text)


class TestAclPlugin(BasePlugin):

    @cmd(admin_only=True)
    def admincmd(self, msg, args):
        """A command only admins should be able to run."""
        return ':boom:'

    @cmd(acl='test')
    def acltest(self, msg, args):
        """A command only members of 'test' should be able to run."""
        return ':sushi:'

    @cmd(admin_only=True, acl='test')
    def adminacl(self, msg, args):
        """Only admins who are in 'test' should be able to run this."""
        return ':godmode:'



from acl import AuthManager
from core import Core
from state.file import FileStateHandler
from user import UserManager

# TODO: Perhaps this one should be moved to a separate package to avoid
# the redis/mysql requirement for those who don't want it.
from state.redis_handler import RedisStateHandler
from state.mysql_handler import MySQLStateHandler



import json
import logging
from datetime import datetime


class PluginManager(object):
    def __init__(self, bot, test_mode=False):
        self.bot = bot
        self.config = bot.config
        self.dispatcher = bot.dispatcher
        self.log = logging.getLogger(__name__)
        self.plugins = []
        self.state_handler = None
        self.test_mode = test_mode

        if self.test_mode:
            self.metrics = {
                'plugins_total': 0,
                'plugins_loaded': 0,
                'load_times': {},
                'plugins_failed': [],
            }

    def load(self):
        import os
        import sys

        # Add plugin dir for extra plugins
        sys.path.append(os.path.join(os.getcwd(), self.config['plugin_dir']))
        if 'plugins' not in self.config:
            self.config['plugins'] = []

        # Add core plugins
        self.config['plugins'].insert(0, 'slackminion.plugins.core.Core')

        for plugin_name in self.config['plugins']:
            if self.test_mode:
                plugin_start_time = datetime.now()
                self.metrics['plugins_total'] += 1
            # module_path.plugin_class_name
            module, name = plugin_name.rsplit('.', 1)
            try:
                plugin = __import__(module, fromlist=['']).__dict__[name]
            except ImportError:
                self.log.exception("Failed to load plugin %s", name)
                if self.test_mode:
                    self.metrics['plugins_failed'].append(name)
                continue

            # load plugin config if available
            config = {}
            if name in self.config['plugin_settings']:
                config = self.config['plugin_settings'][name]
            try:
                p = plugin(self.bot, config=config)
                self.dispatcher.register_plugin(p)
                self.plugins.append(p)
                if p._state_handler:
                    self.state_handler = p
                if self.test_mode:
                    self.metrics['plugins_loaded'] += 1
                    self.metrics['load_times'][name] = (datetime.now() - plugin_start_time).total_seconds() * 1000.0
            except:
                self.log.exception("Failed to register plugin %s", name)
                if self.test_mode:
                    self.metrics['plugins_failed'].append(name)

    def save_state(self):
        if self.state_handler is None:
            self.log.warn("Unable to save state, no handler registered")
            return

        state = {}
        savable_plugins = filter(lambda x: x._dont_save is False, self.plugins)
        for p in savable_plugins:
            attr_blacklist = [
                '_bot',
                '_dont_save',
                '_state_handler',
                '_timer_callbacks',
                'config',
                'log',
            ]
            attrs = {k: v for k, v in p.__dict__.iteritems() if k not in attr_blacklist}
            state[p.__class__.__name__] = attrs
            self.log.debug("Plugin %s: %s", p.__class__.__name__, attrs)
        state = json.dumps(state)
        self.log.debug("Sending the following to the handler: %s", state)
        try:
            self.state_handler.save_state(state)
        except:
            self.log.exception("Handler failed to save state")

    def load_state(self):
        if self.state_handler is None:
            self.log.warn("Unable to load state, no handler registered")
            return

        try:
            state = self.state_handler.load_state()
            state = json.loads(state)
        except:
            self.log.exception("Handler failed to load state")
            return

        for p in self.plugins:
            plugin_name = p.__class__.__name__
            if plugin_name in state:
                self.log.info("Loading state data for %s", plugin_name)
                for k, v in state[plugin_name].iteritems():
                    self.log.debug("%s.%s = %s", plugin_name, k, v)
                    setattr(p, k, v)



import MySQLdb
import MySQLdb.cursors

from datetime import datetime

from . import BaseStateHandler


class MySQLStateHandler(BaseStateHandler):
    def load_state(self):
        conn, cursor = self._get_mysql_info()
        cursor.execute("SELECT id, data FROM state ORDER BY id DESC LIMIT 1")
        res = cursor.fetchone()
        self.log.info("Loaded state id %s", res['id'])
        state = res['data']
        cursor.close()
        return state

    def save_state(self, state):
        conn, cursor = self._get_mysql_info()
        cursor.execute(
            "INSERT INTO state (`timestamp`, `data`) VALUES (%s, %s)",
            (datetime.utcnow(), state)
        )
        conn.commit()

    def _get_mysql_info(self):
        conn = MySQLdb.connect(
            host=self.config['host'],
            user=self.config['user'],
            passwd=self.config['passwd'],
            db=self.config['db'],
            cursorclass=MySQLdb.cursors.DictCursor,
        )
        return conn, conn.cursor()



from slackminion.plugin import cmd
from slackminion.plugin.base import BasePlugin


class AuthManager(BasePlugin):
    """Basic Authorization Plugin"""
    def on_load(self):

        # Hook into the dispatcher to receive acl checks
        setattr(self._bot.dispatcher, 'auth_manager', self)

        # Setup default ACL
        # ACL Rule Ordering (first match)
        # Allow Rule --> Deny Rule --> Allow All
        self._acl = {
            '*': {
                'allow': [],
                'deny': []
            },
        }

        super(AuthManager, self).on_load()

    @cmd(admin_only=True)
    def acl(self, msg, args):
        """ACL Management.

        Usage:
        !acl _action_ [args]

        Actions:
        new _acl_ - Create a new ACL
        delete _acl_ - Delete an ACL

        allow _acl_ _user_ - Add user to the acl allow block
        deny _acl_ _user_ - Add user to the acl deny block
        remove _acl_ _user_ - Remove user from acl allow and deny blocks

        show - Show all defined ACLs
        show _acl_ - Show allow and deny blocks of specified ACL
        """
        if len(args) == 0:
            return "Usage: !acl show OR !acl _action_ _args_"

        action = args[0]
        acl_name = None
        if len(args) == 1:
            # Only show is allowed to have no arguments
            if action != 'show':
                return "Usage: !acl _action_ _args_"
        else:
            acl_name = args[1]

        # Only 'new' can have an acl name that doesn't exist
        # Also exempt 'show' as 'None' will never be in the acl list
        if action not in ['new', 'show'] and acl_name not in self._acl:
            return "ACL %s does not exist" % acl_name

        valid_actions = ['allow', 'deny', 'remove', 'show', 'new', 'delete']
        if action not in valid_actions:
            return "Valid actions: %s" % ', '.join(valid_actions)

        # allow, deny, remove all require two arguments (acl name, username)
        if action in ['allow', 'deny', 'remove'] and len(args) < 3:
            return "Usage: !acl %s %s _args_" % (action, acl_name)

        if action == "allow":
            if self.acl_allow(acl_name, args[2]):
                return "Added %s to %s (allow)" % (args[2], acl_name)
            return "Failed to add %s to %s (allow)" % (args[2], acl_name)

        elif action == "deny":
            if self.acl_deny(acl_name, args[2]):
                return "Added %s to %s (deny)" % (args[2], acl_name)
            return "Failed to add %s to %s (deny)" % (args[2], acl_name)

        elif action == "remove":
            if self.acl_remove(acl_name, args[2]):
                return "Removed %s from %s (allow and deny)" % (args[2], acl_name)
            return "Failed to remove %s from %s (allow and deny)" % (args[2], acl_name)

        elif action == "show":
            output = self.acl_show(msg.user, acl_name)
            if output is None:
                return "Sorry, couldn't find an acl named '%s'" % acl_name
            return output

        elif action == "new":
            if self.acl_new(acl_name):
                return "Created new acl '%s'" % acl_name
            return "ACL '%s' already exists" % acl_name

        elif action == "delete":
            if self.acl_delete(acl_name):
                return "Deleted acl '%s'" % acl_name
            return "ACL '%s' does not exist" % acl_name

    def acl_allow(self, name, user):
        """Add a user to the given acl allow block."""

        # Clear user from both allow and deny before adding
        if not self.acl_remove(name, user):
            return False

        if name not in self._acl:
            return False

        self._acl[name]['allow'].append(user)
        return True

    def acl_deny(self, name, user):
        """Add a user to the given acl deny block."""
        if not self.acl_remove(name, user):
            return False

        if name not in self._acl:
            return False

        self._acl[name]['deny'].append(user)
        return True

    def acl_remove(self, name, user):
        """Remove a user from the given acl (both allow and deny)."""
        if name not in self._acl:
            return False
        if user in self._acl[name]['allow']:
            self._acl[name]['allow'].remove(user)
        if user in self._acl[name]['deny']:
            self._acl[name]['deny'].remove(user)
        return True

    def acl_show(self, user, name=None):
        """Show current allow and deny blocks for the given acl."""
        if name is None:
            return "%s: The following ACLs are defined: %s" % (user, ', '.join(self._acl.keys()))

        if name not in self._acl:
            return None

        return '\n'.join([
            "%s: ACL '%s' is defined as follows:" % (user, name),
            "allow: %s" % ', '.join(self._acl[name]['allow']),
            "deny: %s" % ', '.join(self._acl[name]['deny'])
        ])

    def acl_new(self, name):
        """Create a new acl."""
        if name in self._acl:
            return False

        self._acl[name] = {
            'allow': [],
            'deny': []
        }
        return True

    def acl_delete(self, name):
        """Delete an acl."""
        if name not in self._acl:
            return False

        del self._acl[name]
        return True

    @staticmethod
    def admin_check(cmd, user):
        # Commands not needing admin-level access pass
        if not cmd.admin_only:
            return True

        if hasattr(user, 'is_admin'):
            return user.is_admin
        return False

    def acl_check(self, cmd, user):
        effective_acl = cmd.acl
        if effective_acl not in self._acl:
            self.log.warn("Unable to locate ACL %s for %s, defaulting to *", effective_acl, cmd.method.__name__)
            effective_acl = '*'

        if self._check_allow(self._acl[effective_acl], user):
            return True
        if self._check_deny(self._acl[effective_acl], user):
            return False
        return True

    @staticmethod
    def _check_allow(acl, user):
        return '*' in acl['allow'] or user.username in acl['allow']

    @staticmethod
    def _check_deny(acl, user):
        return '*' in acl['deny'] or user.username in acl['deny']



from base import BasePlugin
from manager import PluginManager


def cmd(admin_only=False, acl='*', aliases=None, *args, **kwargs):
    """
    Decorator to mark plugin functions as commands in the form of !<cmd_name>

    * admin_only - indicates only users in bot_admin are allowed to execute (only used if AuthManager is loaded)
    * acl - indicates which ACL to perform permission checks against (only used if AuthManager is loaded)
    * aliases - register function with additional commands (i.e. !alias1, !alias2, etc)
    """
    def wrapper(func):
        func.is_cmd = True
        func.admin_only = admin_only
        func.acl = acl
        func.aliases = aliases
        return func
    return wrapper


def webhook(*args, **kwargs):
    """
    Decorator to mark plugin functions as entry points for web calls

    * route - web route to register, uses Bottle syntax
    """
    def wrapper(func):
        func.is_webhook = True
        func.route = args[0]
        func.form_params = kwargs['form_params']
        return func
    return wrapper
